#!/usr/bin/env python
#-*- coding:utf-8 -*-
#==============================================================================
#  Copyright (C) 2010 Schaefer & Tobies SuC GmbH.
#  Author: Markus Grimm <mgr@schaefer-tobies.de>
#
#  This file is part of Pyontapi.
#
#  Pyontapi is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyontapi is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with Pyontapi.  If not, see <http://www.gnu.org/licenses/>.
#
#==============================================================================

"""
This module is used to invoke ONTAPI commands on a NetApp Filer, NetCache,
Agent or DFM.

After creating a :class:`schtob.pyontapi.NaFiler` object, it
automatically invokes the API command **system-get-ontapi-version** on the
filer to determine the API version.

Then, if there is no module
*schtob.pyontapi.api.ontapi_<major_version>_<minor_version>*, it generates the
module using some API commands which are available since ONTAPI 1.6.
If you want to know which API commands are available on your filer, read the
original NetApp documentation or instantiate the filer object for the
first time and have a look at the generated python code.

The API commands are wrapped into command packages which are added to the
:class:`schtob.pyontapi.NaFiler` instance as instance attributes.

This example shows you how to use a filer instance::

    >>> from schtob.pyontapi import NaFiler, errors
    >>> filer = NaFiler('filer-name', {'user': 'root', 'password': 'foobar'})
    >>> try:
    ...     result = filer.volume.list_info()
    ... except errors.APIFailure, exc:
    ...     print exc.get_error()
    ...
    >>> print result['volumes']

Another example showing how to get all snapshots for volume vol0:

    >>> from schtob.pyontapi import NaFiler, errors
    >>> filer = NaFiler('filer-name', {'user': 'root', 'password': 'foobar'})
    >>> try:
    ...     result = filer.snapshot.list_info(volume='vol0')
    ... except errors.APIFailure, exc:
    ...     print exc.get_error()
    ...
    >>> print result['snapshots']
    [{'access-time': 1292425201,
      'cumulative-percentage-of-used-blocks': 1,
      'cumulative-total': 6996,
      'name': 'hourly.0',
      'percentage-of-used-blocks': 1,
      'total': 6996}]

A more advanced command for modifying NFS exports:

    >>> from schtob.pyontapi import NaFiler, errors
    >>> filer = NaFiler('filer-name', {'user': 'root', 'password': 'foobar'})
    >>> try:
    ...     result = filer.nfs.exportfs_modify_rule(persistent=False, rule={
    ...         'pathname': '/vol/vol0',
    ...         'read-write': [{'all-hosts':True}],
    ...         'root': [{'name':'my-machine-name'}]
    ...     })
    ...     print "rule successfully modified"
    ... except errors.APIFailure, exc:
    ...     print exc.get_error()
    ...     return False
"""

import base64
import logging
import sys
import xml.dom.minidom

from schtob.pyontapi import api, constants, errors
from schtob.pyontapi.http_responses import httplib

# Custom log level for pyontap logger. set to logging.DEBUG for debugging
# if set to logging.NOTSET, the default logger settings are used
# visit http://docs.python.org/library/logging.html for more details.
LOG_LEVEL = logging.NOTSET


class NaFiler(object):

    """
    Create a new connection to filer ``filer`` using ``settings`` dict.

    ``settings`` may consist of:

    +--------------------+-----------+-------------------------------------+
    | Key                | Default   | Possible values                     |
    +====================+===========+=====================================+
    | **user**           | "root"    | ``str``                             |
    +--------------------+-----------+-------------------------------------+
    | **password**       | ""        | ``str``                             |
    +--------------------+-----------+-------------------------------------+
    | **style**          | ``LOGIN`` | ``LOGIN``, ``HOSTS``                |
    +--------------------+-----------+-------------------------------------+
    | **vfiler**         | ""        | ``str``                             |
    +--------------------+-----------+-------------------------------------+
    | **server_type**    | "Filer"   | "Filer", "NetCache", "DFM", "Agent" |
    +--------------------+-----------+-------------------------------------+
    | **transport_type** | ``HTTP``  | ``HTTP``, ``HTTPS``                 |
    +--------------------+-----------+-------------------------------------+
    | **port**           | ``None``  | ``int``                             |
    +--------------------+-----------+-------------------------------------+
    | **url**            | ``None``  | ``str``                             |
    +--------------------+-----------+-------------------------------------+

    Note that ``ontapi_version`` will be set automatically to your filers
    ontapi version, as pyontapi uses the correct api version for each filer
    instance.
    """

    def __init__(self, filer, settings=None):
        self._filer = filer
        self._log = logging.getLogger('pyontapi')
        self._api_modules = {}

        self._log.setLevel(LOG_LEVEL)

        self._settings = {
            'user': 'root',
            'password': '',
            'style': constants.LOGIN,
            'vfiler': '',
            'server_type': 'Filer',
            'transport_type': constants.HTTP,
            'port': None,
            'url': None,
            'ontapi_version': '1.0',
        }

        if settings and isinstance(settings, dict):
            self.__test_settings(settings)
            self._settings.update(settings)

        self.__handle_servertype()

        if settings and isinstance(settings, dict):
            if not 'transport_type' in settings or \
               settings['transport_type'] == constants.HTTPS:
                # automatically try to use HTTPS if argument not specified or
                # if transport_type is set to HTTPS
                self.__test_https()
        self.__handle_transport_type()
        self.__set_api_classes()

    def __set_api_classes(self):
        """ Add all api classes as class attributes. """

        try:
            result = self.__get_ontapi_version()
        except errors.APIFailure:
            exc = sys.exc_info()[1]
            self._log.error("Cant't get ontapi version. error was: %s",
                        exc.get_error())
            raise errors.UnknownOntapiVersionError(exc.errno,
                    "Cant't get ontapi version. error was:\n%s" % exc.reason
            )

        self._settings['ontapi_version'] = \
            '%(major-version)s.%(minor-version)s' % result

        from schtob.pyontapi import py_gen
        api_modules = py_gen.generate(self)
        for key, value in api_modules.items():
            self._api_modules[key] = api.BaseAPI(self, value)
            setattr(self, key, self._api_modules[key])

    def get_api_modules(self):
        """
        Get all API classes as a dictionary of `package_name`: `api_class`
        """
        return self._api_modules

    def get_api_module(self, package_name):
        """ Get API class for ``package_name``. """
        return self._api_modules.get(package_name, None)

    def do_api_call(self, api_command_name, arguments, fields):
        """
        Create new API call for ``api_command_name`` using ``arguments`` and
        return the result as a dictionary using ``fields`` to parse the
        response.
        """

        xmlcontent = self.__get_xml_content(api_command_name, arguments)
        self._log.debug('XML request: %s', xmlcontent.toprettyxml())

        content = xmlcontent.toxml(encoding='utf-8')

        connection = self.__get_connection()
        connection.putrequest('POST', self._settings['url'])
        connection.putheader('Content-Length', len(content))
        connection.putheader('Content-type', 'text/xml; charset="UTF-8"')

        if self._settings['style'] == constants.LOGIN:
            login = '%(user)s:%(password)s' % self._settings
            if sys.version_info >= (3, 0):
                encoded = base64.encodestring(login.encode()).decode().strip()
            else:
                encoded = base64.encodestring(login)[:-1]
            connection.putheader('Authorization', 'Basic %s' % encoded)

        connection.endheaders()
        connection.send(content)

        try:
            response = connection.getresponse()
        except Exception:
            exc = sys.exc_info()[1]
            raise errors.APIFailure(-1, str(exc))

        if not response.status == 200:
            raise errors.APIFailure(13002,
                'HTTP result status "%s"' % httplib.responses[response.status]
            )

        dom = xml.dom.minidom.parseString(
            response.read().decode('latin1').encode('utf-8')
        )

        connection.close()

        self._log.debug('XML response: %s', dom.toprettyxml())

        return self.__parse_dom(dom, fields)

    def __get_xml_content(self, api_command_name, arguments):
        """
        Create a XML query document for ``api_command_name`` and pass
        ``arguments``.

        ``arguments`` is a list of :class:`schtob.pyontapi.base.NaArgument`
        instances. Each instance contains the argument value and the
        information how to generate the corresponding XML structure.
        """

        doc = xml.dom.minidom.Document()
        netapp = doc.createElement('netapp')

        if self._settings['vfiler']:
            netapp.setAttribute('vfiler', self._settings['vfiler'])
        netapp.setAttribute('version', self._settings['ontapi_version'])
        doc.appendChild(netapp)

        api_element = doc.createElement(api_command_name)
        netapp.appendChild(api_element)

        for arg in arguments:
            arg.append_to_doc(doc, api_element)
        return doc

    def __parse_dom(self, dom, fields):
        """
        Parse XML response ``dom`` using ``fields``.

        ``fields`` is a list of :class:`schtob.pyontap.api.base.NaField`.
        """

        results = dom.getElementsByTagName("results")[0]

        status = results.getAttribute('status')

        if status != 'passed':
            errno = results.getAttribute('errno')
            reason = results.getAttribute('reason')
            raise errors.APIFailure(errno, reason)

        self._log.debug(results.toprettyxml())

        value = {}
        for field in fields:
            value[field.name] = field.get_value(results)
        return value

    def __get_connection(self):
        """ Returns a HTTP/HTTPS connection instance to filer. """

        conn_cls = httplib.HTTPConnection
        if self._settings['transport_type'] == constants.HTTPS:
            conn_cls = httplib.HTTPSConnection

        return conn_cls(self._filer, self._settings['port'])

    def __get_ontapi_version(self):
        """
        Invokes API call `system-get-ontapi-version` to get ONTAPI version.
        """

        from schtob.pyontapi.system import System
        system_api = System(self)
        return system_api.get_ontapi_version()

    #---------------------------PRIVATE METHODS--------------------------------

    def __handle_servertype(self):
        """ Set url and port according to ``self.settings['server_type']``. """

        self._settings['port'] = constants.PORTS[self._settings['server_type']]
        self._settings['url'] = constants.URLS[self._settings['server_type']]

    def __test_https(self):
        """ Test if a HTTPS connection is possible for this filer. """
        import socket

        server_socket = socket.socket()
        server_socket.settimeout(0.25)

        try:
            server_socket.connect((self._filer, 443))
            server_socket.close()
            self._log.debug('HTTPS test was successful')
            self._settings['transport_type'] = constants.HTTPS
        except socket.error:
            if self._settings['transport_type'] == constants.HTTPS:
                msg = "Fall back to HTTP for filer <%s>; HTTPS was specified!"
                self._log.warning(msg, self._filer)
                self._settings['transport_type'] = constants.HTTP

    def __handle_transport_type(self):
        """ Set port according to ``self.settings['transport_type'].`` """

        if self._settings['transport_type'] == constants.HTTPS and \
           self._settings['port'] is None:
            self._settings['port'] = 443

    def __test_settings(self, settings):
        """ Test if settings are valid. """

        test_dict = {
            'style': constants.STYLES,
            'server_type': constants.SERVER_TYPES,
            'transport_type': constants.TRANSPORT_TYPES,
        }

        for key, value_list in test_dict.items():
            # Test if kwargs values are valid
            if key in settings and settings[key] not in value_list:
                msg = "%s is not a valid value for %s" % (settings[key], key)
                self._log.error(msg)
                raise ValueError(msg)
