#!/usr/bin/env python
#-*- coding:utf-8 -*-
#==============================================================================
#  Copyright (C) 2010 Schaefer & Tobies SuC GmbH.
#  Author: Markus Grimm <mgr@schaefer-tobies.de>
#
#  This file is part of Pyontapi.
#
#  Pyontapi is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyontapi is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with Pyontapi.  If not, see <http://www.gnu.org/licenses/>.
#
#==============================================================================

""" module for types generation. """

import sys

from schtob.pyontapi import errors

GENERIC_TYPEDEFS = {
    'boolean': bool,
    'integer': int,
    'string': str,
}

GENERIC_TYPEDEFS_REVERSE = dict([(v, k) for k, v in GENERIC_TYPEDEFS.items()])


class APICommand(object):

    """ Representation of API Command for generating API package. """

    def __init__(self, name, elements):
        self.name = name
        self.elements = elements

    def get_package(self):
        """ get package of this API command. """
        return self.name.split('-')[0]

    def get_arguments(self):
        """ Yields input elements. """
        for element in self.elements:
            if not element.is_output:
                yield element

    def get_output_fields(self):
        """ Yields output elements. """
        for element in self.elements:
            if element.is_output:
                yield element

    def get_py_name(self):
        """ get Python method name for this API command. """
        name = '_'.join(self.name.split('-')[1:])
        if name == 'break' or name[0].isdigit():
            name = '_'.join(self.name.split('-'))
        return name

    def get_py_argument_names(self):
        """
        Getter for Python arguments. list contains required arguments at the
        beginning followed by the default arguments.
        """
        for arg in self.get_required_args():
            yield arg
        for arg in self.get_optional_args():
            yield arg

    def get_optional_args(self):
        """ get optional arguments """
        for arg in self.get_arguments():
            if arg.is_optional:
                yield '``%s`` : %s [default = %s]' % (
                    arg.name_to_py(),
                    arg.get_type_name(),
                    repr(arg.get_default())
                )

    def get_required_args(self):
        """ get required arguments """
        for arg in self.get_arguments():
            if not arg.is_optional:
                yield '``%s`` : %s' % (arg.name_to_py(), arg.get_type_name())

    def get_output_fields_and_types(self):
        """ get output field names and types """
        for field in self.get_output_fields():
            yield '``%s`` : %s' % (field.name, field.get_type_name())


class TypeDef(object):

    """ API Type class. """

    fields = (
        'name', 'type_name', 'encrypted', 'nonempty', 'is_optional',
        'is_output', 'is_validated', 'is_array',
    )

    bools = {
        'true': True,
        'false': False,
        True: True,
        False: False,
        None: False,
    }

    defaults = {
        str: '',
        bool: None,
    }

    def __init__(self, info, var_type):
        self.var_type = var_type
        self.name = info.get('name')
        self.encrypted = info.get('encrypted', False)
        self.nonempty = self.bools[info.get('is-nonempty',  'false')]
        self.is_optional = self.bools[info.get('is-optional',  'false')]
        self.is_output = self.bools[info.get('is-output', 'false')]
        self.is_array = ('[]' in info.get('type'))

    def get_type(self):
        """
        Get type as string. if type is non generic, this returns
        types.get_py_cls().
        """

        if isinstance(self.var_type, NamedType):
            return self.var_type.get_py_cls()
        return self.var_type

    def name_to_py(self):
        """ Returns a pythonic name for this type. """
        return '_'.join(self.name.split('-'))

    def get_default(self):
        """ get default value """
        if self.is_array:
            return None
        return self.defaults.get(self.var_type, None)

    def get_type_name(self):
        """ get type name """
        if isinstance(self.var_type, NamedType):
            var_type = self.var_type.name
        else:
            var_type = GENERIC_TYPEDEFS_REVERSE[self.var_type]
        if self.is_array:
            var_type += '[]'
        return var_type

    def get_py_type_cls(self):
        """
        Returns the Python code of the type class if the type is used in a
        complex type.
        """

        from schtob.pyontapi import api
        return api.NaTypeElement({
            'name': self.name,
            'type': self.get_type(),
            'encrypted': self.encrypted,
            'is-array': self.is_array,
            'nonempty': self.nonempty,
            'is-optional': self.is_optional,
        })

    def get_py_cls(self, value=None):
        """ Returns the Python code for this type. """
        from schtob.pyontapi import api
        if self.is_output:
            return api.NaField({
                'name': self.name,
                'type': self.get_type(),
                'encrypted': self.encrypted,
                'is-array': self.is_array,
            })
        else:
            return api.NaArgument(value, {
                'name': self.name,
                'type': self.get_type(),
                'encrypted': self.encrypted,
                'nonempty': self.nonempty,
                'is-optional': self.is_optional,
                'is-array': self.is_array,
            })


class NamedType(object):

    """ Class for combined (complex) types. """

    def __init__(self, name, elements):
        self.name = name
        self.elements = elements

    def get_py_cls(self):
        """ Returns the Python code for this type. """
        from schtob.pyontapi import api
        elements = [element.get_py_type_cls() for element in self.elements]
        return api.NaNamedType(self.name, elements)

    def duplicate(self):
        """ Clone this type. """
        return NamedType(self.name, self.elements[:])

    def set_output_value_for_childs(self, output):
        """ Sets is_output attribute of all child elements to ``output``. """
        for child in self.elements:
            child.is_output = output


def gen_typedefs(api):
    """ Generate typedef classes. """
    typedefs = {}
    try:
        result = api.get_api_list_types()
    except errors.APIFailure:
        exc = sys.exc_info()[1]
        raise errors.APIGenerationError(exc.errno,
                'cannot get api types! Error was:\n%s' % exc.reason
        )

    for element in result['type-entries']:
        typedefs[element['name']] = NamedType(
            element['name'], element['type-elements']
        )

    for typedef in typedefs.values():
        elements = []
        for element in typedef.elements:
            if element['type'] is None:
                element['type'] = 'string'
            var_type = element['type'].replace('[]', '')
            if var_type in GENERIC_TYPEDEFS:
                var_type = GENERIC_TYPEDEFS[var_type]
            else:
                var_type = typedefs[var_type]
            _element = TypeDef(element, var_type)
            elements.append(_element)
        typedef.elements = elements
    return typedefs
